(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Order_Impl
imports Term_Order
begin

locale pre_redtriple_impl =
  pre_redtriple S NS NST for S NS NST :: "('f :: show, 'v :: show) trs" + 
  fixes checkS :: "('f, 'v) rule \<Rightarrow> shows check"
    and checkNS :: "('f, 'v) rule \<Rightarrow> shows check"
    and checkNST :: "('f, 'v) rule \<Rightarrow> shows check"
  assumes checkSSound: "isOK (checkS rule) \<Longrightarrow> rule \<in> S" 
    and checkNSSound: "isOK (checkNS rule)  \<Longrightarrow> rule \<in> NS"
    and checkNSTSound: "isOK (checkNST rule)  \<Longrightarrow> rule \<in> NST"

locale redtriple_impl = pre_redtriple_impl + redtriple S NS NST 

locale mono_redtriple_impl = redtriple_impl + 
  fixes checkMono :: "shows check"
  assumes checkMonoSound: "isOK (checkMono) \<Longrightarrow> ctxt_closed S"

locale ce_af_redtriple_impl = redtriple_impl + ce_af_redtriple

locale ce_mono_redtriple_impl = mono_ce_redtriple + mono_redtriple_impl

locale mono_ce_af_redtriple_impl = ce_mono_redtriple_impl + ce_af_redtriple_impl

locale mono_ce_af_redtriple_order_impl = redtriple_order + ce_mono_redtriple_impl + ce_af_redtriple_impl

(* valid, check_strict, check_non_strict, check_non_strict_top, af, check_mono, description *)
record ('f, 'v) redtriple =
  valid :: "shows check"
  s :: "('f, 'v) rule \<Rightarrow> shows check"
  ns :: "('f, 'v) rule \<Rightarrow> shows check"
  nst :: "('f, 'v) rule \<Rightarrow> shows check"
  af :: "'f af"
  mono :: "('f, 'v) rule list \<Rightarrow> shows check"
  desc :: "shows"
  cpx :: "('f, 'v) complexity_measure \<Rightarrow> shows + complexity_class"

hide_const (open) valid s ns nst af mono desc cpx

record ('f, 'v) root_redtriple =
  valid :: "shows check"
  s :: "('f, 'v) rule \<Rightarrow> shows check"
  ns :: "('f, 'v) rule \<Rightarrow> shows check"
  nst :: "('f, 'v) rule \<Rightarrow> shows check"
  af :: "'f af"
  aft :: "'f af"
  desc :: "shows"

hide_const (open) valid s ns nst af desc aft

abbreviation (input) no_complexity where
  "no_complexity \<equiv> (\<lambda> _ . Inl (shows ''complexity analysis unsupported'') :: shows + complexity_class)"

interpretation cpx_term_rel S no_complexity 
  by (unfold_locales) auto

locale generic_redtriple_impl = 
  fixes generate_redtriple :: "'a \<Rightarrow> ('f :: show, 'v :: show) redtriple" 
  assumes generate_redtriple:
    "\<lbrakk>isOK (redtriple.valid (generate_redtriple rp));
    isOK (check_allm (redtriple.s (generate_redtriple rp)) s_list);
    isOK (check_allm (redtriple.ns (generate_redtriple rp)) ns_list);
    isOK (check_allm (redtriple.nst (generate_redtriple rp)) nst_list)\<rbrakk> \<Longrightarrow>
    \<exists> S NS NST. 
    cpx_ce_af_redtriple_order S NS NST (redtriple.af (generate_redtriple rp)) (redtriple.cpx (generate_redtriple rp)) \<and> 
    set s_list \<subseteq> S \<and> 
    set ns_list \<subseteq> NS \<and> 
    set nst_list \<subseteq> NST \<and> 
    (isOK (redtriple.mono (generate_redtriple rp) (s_list @ ns_list @ nst_list)) \<longrightarrow>
    mono_ce_af_redtriple_order S NS NST (redtriple.af (generate_redtriple rp)) \<and> ctxt_closed S)"

locale generic_root_redtriple_impl = 
  fixes generate_root_redtriple :: "'a \<Rightarrow> ('f :: show,'v :: show) root_redtriple" 
  assumes generate_root_redtriple:
    "\<lbrakk>isOK (root_redtriple.valid (generate_root_redtriple rp));
    isOK (check_allm (root_redtriple.s (generate_root_redtriple rp)) s_list);
    isOK (check_allm (root_redtriple.ns (generate_root_redtriple rp)) ns_list);
    isOK (check_allm (root_redtriple.nst (generate_root_redtriple rp)) nst_list)\<rbrakk> \<Longrightarrow>
    \<exists> S NS NST. 
    ce_af_root_redtriple_order S NS NST (root_redtriple.af (generate_root_redtriple rp)) (root_redtriple.aft (generate_root_redtriple rp)) \<and> 
    set s_list \<subseteq> S \<and> 
    set ns_list \<subseteq> NS \<and> 
    set nst_list \<subseteq> NST"

end

